import { BinaryFile } from './file';

describe('BinaryFile', () => {
  let bfile: BinaryFile;

  beforeAll(() => {
    const data = new Uint8Array(
      [
        0x49, 0x44, 0x50, 0x4F, 0x06, 0x00, 0x00, 0x00,
        0x9E, 0x3A, 0x97, 0x3E, 0x98, 0xD6, 0xA4, 0x3E,
        0x67, 0xF7, 0x8D, 0x3E, 0xC0, 0xA8, 0x0C, 0xC2,
        0xE2, 0x63, 0x50, 0xC2, 0x06, 0xA9, 0xDF, 0xC1,
        0x68, 0x70, 0x9D, 0x42
      ])
    bfile = new BinaryFile(data.buffer);
  });

  it('should create a file from an array', () => {
    expect(bfile).toBeDefined();
    expect(bfile.buffer).toBeDefined();
  });

  it('should seek to an offset', () => {
    bfile.seek(1);
    // 0x44 = 68 base 10 (or D char)
    const v = bfile.readUInt8();
    expect(v.toString()).toEqual('68');
    expect(String.fromCharCode(v))
      .toEqual(String.fromCharCode(0x44));
  });

  it('should report eof', () => {
    expect(bfile.eof()).toBeFalsy();
  });

  it('should skip bits', () => {
    bfile.seek(0);
    expect(bfile.readInt8()).toEqual(73);
    bfile.seek(3);
    expect(bfile.readInt8()).toEqual(79);
  });

  it('should read a js float value', () => {
    bfile.seek(0);
    expect(bfile.readFloat()).toEqual(3494136064);
  });

  it('should read a js int32 value', () => {
    bfile.seek(0);
    expect(bfile.readUInt32()).toEqual(1330660425);
  });

  it('should read a js int16 value', () => {
    bfile.seek(0);
    expect(bfile.readUInt16()).toEqual(17481);
  });

  it('should read a js int8 value', () => {
    bfile.seek(0);
    expect(bfile.readUInt8()).toEqual(73);
  });

  it('should read a js string', () => {
    bfile.seek(0);
    expect(bfile.readString(4)).toEqual('IDPO');
  });

});
