import { MIDI } from './song';
import { dataBase64 } from '../../static/test_mid_b64';

import { createRun, destroyRun, rewindRun } from './player';
import { Sound, SoundSystem } from '../digital/audio';
import { TimeInSeconds } from './types';
import { BeatInfo, createTempo, PerTickCallback, toSignature } from '../metronome';

////////////////////////////////////////////////////////////////////
const MockSetValueAtTime = (_n1: number, _n2: number) => { };

const MockPanner = {
  orientationX: { setValueAtTime: MockSetValueAtTime },
  orientationY: { setValueAtTime: MockSetValueAtTime },
  orientationZ: { setValueAtTime: MockSetValueAtTime },
  positionX: { setValueAtTime: MockSetValueAtTime },
  positionY: { setValueAtTime: MockSetValueAtTime },
  positionZ: { setValueAtTime: MockSetValueAtTime },
  connect: (_x: unknown) => { },
}
const MockAudioNode = {
  connect: (_x: unknown) => { },
}
const MockGainNode = {
  gain: {
    setValueAtTime: MockSetValueAtTime,
  },
  connect: (_x: unknown) => { },
}

const MockSoundSystem = {
  context: () => ({
    createGain: () => MockGainNode,
    createPanner: () => MockPanner,
  }),
  masterGainNode: () => MockGainNode,
  nodeInput: () => MockAudioNode,
  reboot: () => { },
  getAnalyser: () => MockAudioNode,
  volume: () => 1,
  adjustVolume: (_by: number) => { },
  volumeInDB: () => 10,
  setVolumeInDB: (_decibel: number) => { },
  play: (_sound: Sound) => { },
  stop: (_sound: Sound) => { },
  currentTime: () => 0,
  decodeSound: (_data: ArrayBuffer) => { }, // new AudioBuffer({} as AudioBufferOptions),
  resumePlay: () => { },
  playSound: (_sound: Sound, _time: TimeInSeconds) => { },
  handleStateChange: (_e: Event) => { },
}

const MockPerTickCallback: PerTickCallback = (_beatNumber: number, _time: number, _beatInfo: BeatInfo) => [createTempo(480, 30000), toSignature("4/4")];
////////////////////////////////////////////////////////////////////


describe("MIDI player", () => {
  let mocksong: MIDI;
  let system: SoundSystem;

  beforeAll(() => {
    // @ts-ignore
    system = MockSoundSystem;
    const arrayBuffer = Uint8Array.from(window.atob(dataBase64), c => c.charCodeAt(0));
    mocksong = new MIDI(arrayBuffer.buffer);
  });

  it("should be able to create a run", () => {
    const run = createRun(system, mocksong);
    expect(run.tracks.length).toEqual(mocksong.numberOfTracks);
    expect(run.output).not.toBeUndefined();
  });

  it("should be able to assign a tickListener", () => {
    const run = createRun(system, mocksong);
    run.tickListeners.push(MockPerTickCallback);

    expect(run.tickListeners.length).toEqual(1);
  });

  it("should destroy a run", () => {
    const run = createRun(system, mocksong);
    run.tickListeners.push(MockPerTickCallback);

    const disconnectMock = jest.fn();
    run.output.disconnect = disconnectMock;

    destroyRun(system, run);

    expect(disconnectMock).toHaveBeenCalledTimes(1);
    expect(run.tickListeners.length).toEqual(0);
  });

  it("should be able to rewind a run", () => {
    const run = createRun(system, mocksong);
    rewindRun(run);

    expect(run.tracks[0].idx).toEqual(0);
  });

});