export type Effect = {
  command: number;
  argX: number;
  argY: number;
}

// 0 - 127
export type MidiInputNote = number;

/**
 * A float number in seconds. e.g.: 
 * 
 * 1 = 1 second
 * 
 * .5 = 1/2 a second
 * 
 * .001 = 1 millisecond
 * 
 * .01 = 10 millisecond 
 */
export type TimeInSeconds = number;

// TODO: this is gross
/** Look up a note (e.g. G#3), by midi value (e.g. 56) */
export function noteByMidi(value: number): string {
  // @ts-ignore
  const midiVals = [...NoteToMidi.entries()]
    .filter(({ 1: v }) => v === value)
    .map(([k]) => k);
  return midiVals[0];
}

/** Table that maps note values e.g. "G#3" to midi values e.g. "56" */
export const NoteToMidi = new Map<string, number>([
  ["G-9", 127],
  ["Gb9", 126],
  ["F#9", 126],
  ["F-9", 125],
  ["E-9", 124],
  ["Eb9", 123],
  ["D#9", 123],
  ["D-9", 122],
  ["Db9", 121],
  ["C#9", 121],
  ["C-9", 120],
  ["B-9", 119],
  ["Bb8", 118],
  ["A#8", 118],
  ["A-8", 117],
  ["Ab8", 116],
  ["G#8", 116],
  ["G-8", 115],
  ["Gb8", 114],
  ["F#8", 114],
  ["F-8", 113],
  ["E-8", 112],
  ["Eb8", 111],
  ["D#8", 111],
  ["D-8", 110],
  ["Db8", 109],
  ["C#8", 109],
  ["C-8", 108],
  ["B-7", 107],
  ["Bb7", 106],
  ["A#7", 106],
  ["A-7", 105],
  ["Ab7", 104],
  ["G#7", 104],
  ["G-7", 103],
  ["Gb7", 102],
  ["F#7", 102],
  ["F-7", 101],
  ["E-7", 100],
  ["Eb7", 99],
  ["D#7", 99],
  ["D-7", 98],
  ["Db7", 97],
  ["C#7", 97],
  ["C-7", 96],
  ["B-6", 95],
  ["Bb6", 94],
  ["A#6", 94],
  ["A-6", 93],
  ["Ab6", 92],
  ["G#6", 92],
  ["G-6", 91],
  ["Gb6", 90],
  ["F#6", 90],
  ["F-6", 89],
  ["E-6", 88],
  ["Eb6", 87],
  ["D#6", 87],
  ["D-6", 86],
  ["Db6", 85],
  ["C#6", 85],
  ["C-6", 84],
  ["B-5", 83],
  ["Bb5", 82],
  ["A#5", 82],
  ["A-5", 81],
  ["Ab5", 80],
  ["G#5", 80],
  ["G-5", 79],
  ["Gb5", 78],
  ["F#5", 78],
  ["F-5", 77],
  ["E-5", 76],
  ["Eb5", 75],
  ["D#5", 75],
  ["D-5", 74],
  ["Db5", 73],
  ["C#5", 73],
  ["C-5", 72],
  ["B-4", 71],
  ["Bb4", 70],
  ["A#4", 70],
  ["A-4", 69],
  ["Ab4", 68],
  ["G#4", 68],
  ["G-4", 67],
  ["Gb4", 66],
  ["F#4", 66],
  ["F-4", 65],
  ["E-4", 64],
  ["Eb4", 63],
  ["D#4", 63],
  ["D-4", 62],
  ["Db4", 61],
  ["C#4", 61],
  ["C-4", 60],
  ["B-3", 59],
  ["Bb3", 58],
  ["A#3", 58],
  ["A-3", 57],
  ["Ab3", 56],
  ["G#3", 56],
  ["G-3", 55],
  ["Gb3", 54],
  ["F#3", 54],
  ["F-3", 53],
  ["E-3", 52],
  ["Eb3", 51],
  ["D#3", 51],
  ["D-3", 50],
  ["Db3", 49],
  ["C#3", 49],
  ["C-3", 48],
  ["B-2", 47],
  ["Bb2", 46],
  ["A#2", 46],
  ["A-2", 45],
  ["Ab2", 44],
  ["G#2", 44],
  ["G-2", 43],
  ["Gb2", 42],
  ["F#2", 42],
  ["F-2", 41],
  ["E-2", 40],
  ["Eb2", 39],
  ["D#2", 39],
  ["D-2", 38],
  ["Db2", 37],
  ["C#2", 37],
  ["C-2", 36],
  ["B-1", 35],
  ["Bb1", 34],
  ["A#1", 34],
  ["A-1", 33],
  ["Ab1", 32],
  ["G#1", 32],
  ["G-1", 31],
  ["Gb1", 30],
  ["F#1", 30],
  ["F-1", 29],
  ["E-1", 28],
  ["Eb1", 27],
  ["D#1", 27],
  ["D-1", 26],
  ["Db1", 25],
  ["C#1", 25],
  ["C-1", 24],
  ["B-0", 23],
  ["Bb0", 22],
  ["A#0", 22],
]);

/**
 * An array of frequency values based on midi numbers. Using a midi
 * value as an index, this will give you the frequency of that midi
 * note for use in a sine wave (for example)
 */
export const MidiToFreq: Array<number> = [
  8.18,
  8.66,
  9.18,
  9.72,
  10.30,
  10.91,
  11.56,
  12.25,
  12.98,
  13.75,
  14.57,
  15.43,
  16.35,
  17.32,
  18.35,
  19.45,
  20.60,
  21.83,
  23.12,
  24.50,
  25.96,
  27.50,
  29.14,
  30.87,
  32.70,
  34.65,
  36.71,
  38.89,
  41.20,
  43.65,
  46.25,
  49.00,
  51.91,
  55.00,
  58.27,
  61.74,
  65.41,
  69.30,
  73.42,
  77.78,
  82.41,
  87.31,
  92.50,
  98.00,
  103.83,
  110.00,
  116.54,
  123.47,
  130.81,
  138.59,
  146.83,
  155.56,
  164.81,
  174.61,
  185.00,
  196.00,
  207.65,
  220.00,
  233.08,
  246.94,
  261.63,
  277.18,
  293.66,
  311.13,
  329.63,
  349.23,
  369.99,
  392.00,
  415.30,
  440.00,
  466.16,
  493.88,
  523.25,
  554.37,
  587.33,
  622.25,
  659.26,
  698.46,
  739.99,
  783.99,
  830.61,
  880.00,
  932.33,
  987.77,
  1046.50,
  1108.73,
  1174.66,
  1244.51,
  1318.51,
  1396.91,
  1479.98,
  1567.98,
  1661.22,
  1760.00,
  1864.66,
  1975.53,
  2093.00,
  2217.46,
  2349.32,
  2489.02,
  2637.02,
  2793.83,
  2959.96,
  3135.96,
  3322.44,
  3520.00,
  3729.31,
  3951.07,
  4186.01,
  4434.92,
  4698.64,
  4978.03,
  5274.04,
  5587.65,
  5919.91,
  6271.93,
  6644.88,
  7040.00,
  7458.62,
  7902.13,
  8372.02,
  8869.84,
  9397.27,
  9956.06,
  10548.08,
  11175.30,
  11839.82,
  12543.85,
];

/**
 * Different types of oscillators. Note "custom" currently means
 * white noise
 */
export type WaveTypes = "sine" | "square" | "sawtooth" | "triangle" | "custom";

/** Oscillator */
export type VCO = {
  type: WaveTypes,
  frequency?: number,
  freqOffset?: number,
  freqMult?: number,
  detune: number,
}

/** Voltage Control Amplifier */
export type VCA = {
  /** The gain value in absolute terms. 0-1 */
  gain?: number,
  /** Used in additive synths - this is the value difference
   * from the base wave. This is almost always negative. */
  gainOffset?: number,
}

/** Attack, Decay, Sustain, Release */
export type ADSR = {
  /**
   * Percent length of the sample
   * that should be used for attack
   */
  attack: number,
  /**
   * Percent length of the sample that
   * should be used for decay
   */
  decay: number,
  /** 
   * Volume as you are "holding the key down"
   * the volume between decay and release
   */
  sustain: number,
  /** Percent of the sample that should be
   * used for the "fade out" */
  release: number,
}


// https://developer.mozilla.org/en-US/docs/Web/API/BiquadFilterNode
/** High / Low pass filters */
export type Filter = {
  type: 'lowpass' | 'highpass' | 'bandpass' | 'lowshelf' | 'highshelf' | 'peaking' | 'notch' | 'allpass'
  frequency: number,
  detune?: number,
  /**
   * The "quality control" is the width of the frequency range
   * that this filter is going to effect. A reasonable setting is
   * between 1.0 and 4.0
   * 
   * 10.0 is super specific (about 50 hrtz wide)
   * 
   * 1.5 is about one octave
   * 
   * 0.5 is really wide and probably not useful (0 - 10,000 hrtz)
   */
  q: number,
  gain: number,
}

// https://developer.mozilla.org/en-US/docs/Web/API/DynamicsCompressorNode
/** Compressor */
export type Compressor = {
  threshold: number,
  knee: number,
  ratio: number,
  attack: TimeInSeconds,
  release: TimeInSeconds,
}

//////////////////////////////////////

/**
 * A single, simple wave form with dynamic elements. This will only have
 * one sine/square/etc wave. */
export type Voice = {
  vco: VCO,
  vca: VCA,
}

/** 
 * Similar to a Voice, but can have a number of waves added together
 * all these play at the same time - an "additive synth". *
 * Watch out for your ears as adding waves will increase the volume!
 */
export type Patch = {
  /** How long the sound should last for before it fades out naturally.
   * (in seconds). For example a guitar pluck would last longer than
   * a trumpet blast regardless of filters or anything */
  len: TimeInSeconds,
  /** The overall volume of the patch. Between 0 and 1. If you don't
   * specify, default gain will be 1 (100%) */
  vca?: VCA,
  /** Limited to max 8 currently */
  voices: Array<Voice>,
  filter?: Filter,
  compressor?: Compressor,
  adsr?: ADSR,
}

//////////////////////////////////////

// https://en.wikipedia.org/wiki/General_MIDI

// https://soundprogramming.net/file-formats/general-midi-drum-note-numbers/
/** When on channel 10, these are the drum samples that should play for the
 * midi note values. These map 1:1 */
export enum GeneralMIDIPercussion {
  // GM2
  HighQ = 27,
  Slap,
  ScratchPush,
  ScratchPull,
  Sticks,
  SquareClick,
  MetronomeClick,
  MetronomeBell,
  // 35
  BassDrum2,
  BassDrum1,
  SideStick,
  SnareDrum1, // Acoustic Snare
  // 40
  HandClap,
  SnareDrum2,
  LowTom2,
  ClosedHiHat,
  LowTom1,
  // 45
  PedalHiHat,
  MidTom2,
  OpenHiHat,
  MidTom1,
  HighTom2,
  // 50
  CrashCymbal1,
  HighTom1,
  RideCymbal1,
  ChineseCymbal,
  RideBell,
  Tambourine,
  SplashCymbal,
  Cowbell,
  CrashCymbal2,
  VibraSlap,
  RideCymbal2,
  HighBongo,
  LowBongo,
  MuteHighConga,
  OpenHighConga,
  LowConga,
  HighTimbale,
  LowTimbale,
  HighAgogo,
  LowAgogo,
  Cabasa,
  Maracas,
  ShortWhistle,
  LongWhistle,
  ShortGuiro,
  LongGuiro,
  Claves,
  HighWoodBlock,
  LowWoodBlock,
  MuteCuica,
  OpenCuica,
  MuteTriangle,
  OpenTriangle,
  // GM2 (82)
  Shaker,
  JingleBell,
  Belltree,
  Castanets,
  MuteSurdo,
  OpenSurdo
};

// https://www.cs.cmu.edu/~music/cmsip/readings/GMSpecs_Patches.htm
/** When a midi file references an instrument by number, this is what 
 * they mean */
export enum GeneralMidiInstruments {
  // 1-8 PIANO
  AcousticGrand = 1,
  BrightAcoustic,
  ElectricGrand,
  HonkyTonk,
  ElectricPiano1,
  ElectricPiano2,
  Harpsichord,
  Clav,
  // 9-16 CHROMATIC PERCUSSION
  Celesta,
  Glockenspiel,
  MusicBox,
  Vibraphone,
  Marimba,
  Xylophone,
  TubularBells,
  Dulcimer,
  // 17-24 ORGAN
  DrawbarOrgan,
  PercussiveOrgan,
  RockOrgan,
  ChurchOrgan,
  ReedOrgan,
  Accoridan,
  Harmonica,
  TangoAccordian,
  // 25-32 GUITAR
  AcousticGuitar_nylon,
  AcousticGuitar_steel,
  ElectricGuitar_jazz,
  ElectricGuitar_clean,
  ElectricGuitar_muted,
  OverdrivenGuitar,
  DistortionGuitar,
  GuitarHarmonics,
  // 33-40 BASS
  AcousticBass,
  ElectricBass_finger,
  ElectricBass_pick,
  FretlessBass,
  SlapBass1,
  SlapBass2,
  SynthBass1,
  SynthBass2,
  // 41-48 STRINGS
  Violin,
  Viola,
  Cello,
  Contrabass,
  TremoloStrings,
  PizzicatoStrings,
  OrchestralStrings,
  Timpani,
  // 49-56 ENSEMBLE
  StringEnsemble1,
  StringEnsemble2,
  SynthStrings1,
  SynthStrings2,
  ChoirAahs,
  VoiceOohs,
  SynthVoice,
  OrchestraHit,
  // 57-64 BRASS
  Trumpet,
  Trombone,
  Tuba,
  MutedTrumpet,
  FrenchHorn,
  BrassSection,
  SynthBrass1,
  SynthBrass,
  // 65-72 REED
  SopranoSax,
  AltoSax,
  TenorSax,
  BaritoneSax,
  Oboe,
  EnglishHorn,
  Bassoon,
  Clarinet,
  // 73-80 PIPE
  Piccolo,
  Flute,
  Recorder,
  PanFlute,
  BlownBottle,
  Shakuhachi,
  Whistle,
  Ocarina,
  // 81-88 SYNTH LEAD
  Lead1_square,
  Lead2_sawtooth,
  Lead3_calliope,
  Lead4_chiff,
  Lead5_charang,
  Lead6_voice,
  Lead7_fifths,
  Lead8_bass_lead,
  // 89-96 SYNTH PAD
  Pad1_new_age,
  Pad2_warm,
  Pad3_polysynth,
  Pad4_choir,
  Pad5_bowed,
  Pad6_metallic,
  Pad7_halo,
  Pad8_sweep,
  // 97-104 SYNTH EFFECTS
  FX1_rain,
  FX2_soundtrack,
  FX3_crystal,
  FX4_atmosphere,
  FX5_brightness,
  FX6_goblins,
  FX7_echoes,
  FX8_sci_fi,
  // 105-112 ETHNIC
  Sitar,
  Banjo,
  Shamisen,
  Koto,
  Kalimba,
  Bagpipe,
  Fiddle,
  Shanai,
  // 113-120 PERCUSSIVE
  TinkleBell,
  Agogo,
  SteelDrums,
  Woodblock,
  TaikoDrum,
  MelodicTom,
  SynthDrum,
  ReverseCymbal,
  // 121-128 SOUND EFFECTS
  GuitarFretNoise,
  BreathNoise,
  Seashore,
  BirdTweet,
  TelephoneRing,
  Helicopter,
  Applause,
  Gunshot,
};
