import { MOD } from './song';
// @ts-ignore
import { dataBase64 } from '../../static/test_mod_b64';

// audio/x-mod
// const file = new File([arrayBuffer], "dummy.mod", { type: 'audio/mod' });

describe("MOD file format", () => {
  let m: MOD;

  beforeAll(() => {
    const arrayBuffer = Uint8Array.from(window.atob(dataBase64), c => c.charCodeAt(0));
    m = new MOD(arrayBuffer.buffer);
  })

  it("should be able get the song", () => {
    const s = m.song();
    expect(s.positions.length).toBe(128);
    expect(s.length).toBe(2);
    expect(s.type).toBe("M.K.");
    expect(s.title).toBe("Test Song");
  });

  it("should be able to parse sample metadata", () => {
    const sam = m.sample(0);

    expect(sam.name).toBe("Inst 01");
    expect(sam.length).toBe(0x4182);
    expect(sam.repeatPoint).toBe(0x208E);
    expect(sam.repeatLength).toBe(0x20F4);
  });

  it("should be able to parse sample metadata 2", () => {
    const sam = m.sample(1);

    expect(sam.name).toBe("Inst 02");
    expect(sam.length).toBe(0x5D3C);
    expect(sam.repeatPoint).toBe(0x248C);
    expect(sam.repeatLength).toBe(0x38B0);
  });

  it("should be able read a the patterns", () => {
    // We know there are 2 patterns because we made it
    // you'll have to look in song.positions to know in
    // the future
    const ps = m.patterns(2);

    // Bummer - we can't play C-3 values they are too low
    // for the sample rate, but mod seems to know that and
    // set's C-3 values as C-4. Makes composing odd as 
    // everything will be an octave lower in milkytracker
    // https://developer.mozilla.org/en-US/docs/Web/API/AudioContext/AudioContext

    // pattern / row / channel
    expect(ps[0][0][0].note).toEqual("C-4");
    expect(ps[0][0][0].sample).toEqual(1);

    expect(ps[0][2][1].note).toEqual("C-5");
    expect(ps[0][2][1].sample).toEqual(0x16);
  });
});